#include "ui.h"

#include <bits/stdc++.h>

#include "internal.h"

namespace kek {

Font *UIElement::defaultFont = nullptr;

UIElement::UIElement(UIValue x, UIValue y): x(x), y(y) {

}

UIElement::~UIElement() {

}

// Returns the element's position relative to its parent in pixels
UIPoint UIElement::getPosition() {
	return UIPoint(uiToScreen(x), uiToScreen(y));
}

// Returns the element's position on the screen in pixels
UIPoint UIElement::getScreenPosition() {
	UIPoint pos = getPosition();
	if(parent) {
		UIPoint parentPos = parent->getPosition();
		pos.x += parentPos.x;
		pos.y += parentPos.y;
	}
	return pos;
}

void UIElement::addChild(UIElement *child) {
	children.push_back(child);
}

void UIElement::removeChild(UIElement *child) {
	std::remove(children.begin(), children.end(), child);
}

inline int UIElement::uiToScreen(UIValue val) {
	float px = 0;
	px += val.pixels;
	if(parent) px += (int) (val.parentWidth * parent->getBounds().w);
	px += val.screenWidth * kekData.screenWidth;
	if(!parent) px += val.parentWidth * kekData.screenWidth;
	if(parent) px += (int) (val.parentHeight * parent->getBounds().h);
	px += val.screenHeight * kekData.screenHeight;
	if(!parent) px += val.parentHeight * kekData.screenHeight;
	return px;
}

void UIElement::drawAll(UIPoint screenPos, glm::mat4 projection) {
	if(!visible) return;
	draw(screenPos, projection);

	for(UIElement *child : children) {
		UIPoint pos = child->getPosition();
		child->drawAll(UIPoint(screenPos.x + pos.x, screenPos.y + pos.y), projection);
	}
}

void UIElement::hoverAll(UIPoint pos, UIPoint screenPos) {
	UIBounds bounds = getBounds();
	if(!bounds.contains(pos)) { // Only used by topmost parent UIElement
		if(hovering) hoverExitAll();
		return;
	}

	UIElement *hoveredChild = NULL;
	for(UIElement *child : children) {
		UIPoint childPos = child->getPosition();
		int relX = pos.x - childPos.x;
		int relY = pos.y - childPos.y;
		UIBounds b = child->getBounds();
		if(!b.contains(UIPoint(relX, relY))) continue;
		hoveredChild = child;
	}

	for(UIElement *child : children) {
		if(child != hoveredChild && child->hovering) child->hoverExitAll();
	}

	if(hoveredChild) {
		UIPoint childPos = hoveredChild->getPosition();
		hoveredChild->hoverAll(UIPoint(pos.x - childPos.x, pos.y - childPos.y), screenPos);
	}

	if(!hovering) hoverEnter(pos, screenPos);
	hovering = true;
	hover(pos, screenPos);
}

void UIElement::hoverExitAll() {
	if(!hovering) return;
	hovering = false;
	for(UIElement *child : children) {
		if(child->hovering) child->hoverExitAll();
	}
	hoverExit();
}

void UIElement::clickAll(UIPoint pos, UIPoint screenPos, GLFWMouseButton button) {
	UIBounds bounds = getBounds();
	if(!bounds.contains(pos)) return; // Only used by topmost parent UIElement

	UIElement *clickedChild = NULL;
	for(UIElement *child : children) {
		if(!child->clickable) continue;
		UIPoint childPos = child->getPosition();
		int relX = pos.x - childPos.x;
		int relY = pos.y - childPos.y;
		UIBounds b = child->getBounds();
		if(!b.contains(UIPoint(relX, relY))) continue;
		clickedChild = child;
	}

	if(clickedChild != NULL) {
		UIPoint childPos = clickedChild->getPosition();
		clickedChild->clickAll(UIPoint(pos.x - childPos.x, pos.y - childPos.y), screenPos, button);
	}else {
		click(pos, screenPos, button);
	}
}

bool UIElement::focusEnterAll(UIPoint pos, UIPoint screenPos) {
	UIBounds bounds = getBounds();
	if(!bounds.contains(pos)) return false;

	UIElement *focusedChild = nullptr;
	for(UIElement *child : children) {
		UIPoint childPos = child->getPosition();
		int relX = pos.x - childPos.x;
		int relY = pos.y - childPos.y;
		UIBounds b = child->getBounds();
		if(!b.contains(UIPoint(relX, relY))) continue;
		focusedChild = child;
	}

	if(focusedChild) {
		UIPoint childPos = focusedChild->getPosition();
		bool focusable = focusedChild->focusEnterAll(UIPoint(pos.x - childPos.x, pos.y - childPos.y), screenPos);
		if(focusable) return true;
	}

	if(focusable) {
		//UI::focusElement(this); FIXME
		return true;
	}

	return false;
}

UIElement *UIElement::dragEnterAll(UIPoint pos, UIPoint screenPos) {
	UIBounds bounds = getBounds();
	if(!bounds.contains(pos)) return nullptr; // Only used by topmost parent UIElement

	UIElement *draggedChild = nullptr;
	for(UIElement *child : children) {
		UIPoint childPos = child->getPosition();
		int relX = pos.x - childPos.x;
		int relY = pos.y - childPos.y;
		UIBounds b = child->getBounds();
		if(!b.contains(UIPoint(relX, relY))) continue;
		draggedChild = child;
	}

	if(draggedChild) {
		UIPoint childPos = draggedChild->getPosition();
		UIElement *dragged = draggedChild->dragEnterAll(UIPoint(pos.x - childPos.x, pos.y - childPos.y), screenPos);
		if(dragged) return dragged;
	}

	if(draggable) return this;

	return nullptr;
}

void UIElement::init() {
	defaultFont = new Font(KEK_DEFAULT_FONT);
}

}
